# Copyright (C) 2024 Matthias Nadig

from ._config import Config
from ._model import Model

from ._creation_api import create_model, create_config
from ._creation_api import create_from_config
from ._creation_api import add_module

from ._modules import templates



# Copyright (C) 2024 Matthias Nadig

from ._datastructs import as_sample, as_sample_with_channels
from ._datastructs import as_batch, as_batch_with_channels
from ._datastructs import as_array
from ._datastructs import as_label
from ._datastructs import as_bounds
from ._datastructs import as_semantic_map_overlapping, as_semantic_labelmap
from ._datastructs import as_instance_map_overlapping, as_instance_labelmap
from ._datastructs import Sample
from ._datastructs import Batch, BatchSingleTensor, BatchFlexShape
from ._datastructs import BatchChannelMatrix, BatchChannelsSingleTensor, BatchChannelsFlexGrid
from ._datastructs import Array
from ._datastructs import Annotation, Label, Bounds
from ._datastructs import SegmentationMap, SegmentationMapMultiClass
from ._datastructs import SemanticMapOverlapping, SemanticLabelmap
from ._datastructs import InstanceMapOverlapping, InstanceLabelmap
from ._datastructs import raise_if_not_none_or_iterable_of_annotations

from ._creation import create_model, create_config
from ._creation import create_from_config
from ._creation import templates, add_module

from ._io import save_model, load_model
from ._io import save_config, load_config
from ._io import save_checkpoint, load_checkpoint
from ._io import to_storable_dict, from_storable_dict

# Finally import extension after completed initialization of tacotree
import tacotree_lab as lab


__version__ = '1.0.0'
__author__ = 'Matthias Nadig'



# Copyright (C) 2024 Matthias Nadig

import numpy as np
import torch

from ._base_data import BaseData
from ._sample import as_sample, Sample
from ._annotations import raise_if_not_none_or_iterable_of_annotations


def as_batch(samples, annotations=None, limits=None):
    return _as_batch(constructor_static=BatchSingleTensor, constructor_flex=BatchFlexShape, samples=samples, annotations=annotations, limits=limits)


def as_batch_with_channels(samples, annotations=None, limits=None):
    return _as_batch(constructor_static=BatchChannelsSingleTensor, constructor_flex=BatchChannelsFlexGrid, samples=samples, annotations=annotations, limits=limits)


def _as_batch(constructor_static, constructor_flex, samples, annotations=None, limits=None):
    if not hasattr(samples, '__len__'):
        raise TypeError(samples)

    # Make sure annotations are of correct type
    if annotations is not None:
        if not hasattr(annotations, '__len__'):
            raise TypeError(f'Expected iterable with items of Annotation-type, got {annotations}')
        if len(annotations) != len(samples):
            raise ValueError(f'Expected at least one annotation for each sample, got ({len(annotations)} != {len(samples)}) --> {annotations}')
        for annotation_of_sample in annotations:
            raise_if_not_none_or_iterable_of_annotations(annotation_of_sample)

    # Make sure inputs are of correct type
    if isinstance(samples, torch.Tensor):
        batch = _from_torch(constructor=constructor_static, samples=samples, annotations=annotations, limits=limits)
    elif isinstance(samples, np.ndarray):
        batch = _from_numpy(constructor=constructor_static, samples=samples, annotations=annotations, limits=limits)
    elif all([isinstance(s, Sample) for s in samples]):
        # Use annotations from samples
        if annotations is not None:
            raise ValueError
        else:
            annotations = [s.annotations() for s in samples]

        # Expecting samples to have same limits, if no limit is given as argument
        limits_found = [None if s.limits() is None else tuple(s.limits()) for s in samples]
        if limits is None:
            limits = limits_found[0]
        else:
            if not all([limits_found[0] == limits_current for limits_current in limits_found]):
                raise ValueError(
                    f'If not limit is set explicitly, all samples must have same limits. Got: {limits_found}')
        batch = _from_samples(constructor_static=constructor_static, constructor_flex=constructor_flex, samples=samples, annotations=annotations, limits=limits)
    elif all([isinstance(s, np.ndarray) for s in samples]) or all([isinstance(s, torch.Tensor) for s in samples]):
        if annotations is not None and not (hasattr(annotations, '__iter__') and len(annotations) == len(samples)):
            raise ValueError(annotations)
        if limits is not None and not (hasattr(limits, '__iter__') and len(limits) == 2):
            raise ValueError(limits)
        n_dims = len(samples[0].shape) - 1
        if not all([len(s.shape) - 1 == n_dims for s in samples]):
            raise ValueError([s.shape for s in samples])
        samples = [
            as_sample(data=s, n_dims=n_dims) for i, s in enumerate(samples)
        ]
        batch = _from_samples(constructor_static=constructor_static, constructor_flex=constructor_flex, samples=samples, annotations=annotations, limits=limits)
    else:
        types_each_sample = [type(s) for s in samples]
        raise TypeError(f'Expected Numpy array or list with items of Sample-type, got {types_each_sample}')

    return batch


def _from_torch(constructor, samples, annotations, limits):
    return constructor(data=samples, annotations=annotations, limits=limits)


def _from_numpy(constructor, samples, annotations, limits):
    samples = torch.from_numpy(samples)
    return constructor(data=samples, annotations=annotations, limits=limits)


def _from_samples(constructor_static, constructor_flex, samples, annotations, limits):
    # Expecting samples to be of same shapes
    shape_expected = samples[0].data().shape
    is_shape_equal = all([s.data().shape == shape_expected for s in samples])
    if is_shape_equal:
        data = torch.stack([s.data() for s in samples])
        batch = constructor_static(data=data, annotations=annotations, limits=limits)
    else:
        # shape_each_sample = [s.data().shape for s in samples]
        # raise ValueError(f'Expected all shapes to be equal, got shapes {shape_each_sample}')
        # warnings.warn('Batch with flexible shape is created. This should only happen during debugging.')
        data = [s.data() for s in samples]
        batch = constructor_flex(data=data, annotations=annotations, limits=limits)

    return batch


class Batch(BaseData):
    def __init__(self, data, annotations, limits, sample_shape):
        super().__init__(data=data, annotations=annotations)
        self._shape_sample = sample_shape
        self._limits = limits
        self._n_axes = len(self._shape_sample)
        self._n_samples = len(self._data)

    def n_samples(self):
        return self._n_samples

    def n_axes(self):
        return self._n_axes

    def sample_shape(self):
        return self._shape_sample

    def limits(self):
        return self._limits

    def set_limits(self, limits):
        self._limits = limits
        return self


class BatchSingleTensor(Batch):
    def __init__(self, data, annotations=None, limits=None):
        # n_channels = None if len(data.shape) == 1 else data.shape[1]
        sample_shape = data.shape[1:]
        super().__init__(data=data, annotations=annotations, limits=limits, sample_shape=sample_shape)  # , n_channels=n_channels)

    def set_data(self, data):
        """ CAVEAT! Should only be used if modified data has same properties """
        if data.shape != self._data.shape or data.dtype != self._data.dtype:
            raise ValueError(f'Trying to set not matching data (old/new): {self._data.shape} != {data.shape}, {self._data.dtype} != {data.dtype}')
        self._data = data
        return self


class BatchFlexShape(Batch):
    def __init__(self, data, annotations=None, limits=None):
        # n_dims = len(data[0].shape) - 1
        # n_channels = None if len(data[0].shape) == 0 else data[0].shape[0]
        # for sample in data:
        #     if len(sample.shape) - 1 != n_dims or (None if len(sample.shape) == 0 else sample.shape[0]) != n_channels:
        #         raise ValueError([s.shape for s in data])
        # n_dims = max(0, n_dims)  # could be -1 if sample is a scalar
        sample_shape = _get_flexible_shape(data)
        super().__init__(data=data, annotations=annotations, limits=limits, sample_shape=sample_shape)

    def set_data(self, data):
        """ CAVEAT! Should only be used if modified data has same properties """
        raise RuntimeError('Data modification not implemented yet')

    def as_type(self, dtype):
        self._data = [sample.type(dtype) for sample in self._data]
        return self

    def to(self, *args, **kwargs):
        self._data = [sample.to(*args, **kwargs) for sample in self._data]
        return self


class BatchChannelMatrix:
    def __init__(self, n_channels, grid_shape, **kwargs):
        super().__init__(**kwargs)  # required for multi-inheritance
        self._grid_shape = grid_shape
        self._n_dims = len(self._grid_shape)
        self._n_channels = n_channels

    def n_channels(self):
        return self._n_channels

    def n_dims(self):
        return self._n_dims

    def grid_shape(self):
        return self._grid_shape


class BatchChannelsSingleTensor(BatchChannelMatrix, BatchSingleTensor):
    def __init__(self, data, annotations=None, limits=None):
        grid_shape = data.shape[2:]
        # n_channels = None if len(data.shape) == 1 else data.shape[1]
        n_channels = data.shape[1]
        super().__init__(
            n_channels=n_channels, grid_shape=grid_shape,
            data=data, annotations=annotations, limits=limits)


class BatchChannelsFlexGrid(BatchChannelMatrix, BatchFlexShape):
    def __init__(self, data, annotations=None, limits=None):
        sample_shape = _get_flexible_shape(data)
        n_channels = sample_shape[0]
        grid_shape = sample_shape[1:]
        super().__init__(
            n_channels=n_channels, grid_shape=grid_shape,
            data=data, annotations=annotations, limits=limits)


def _get_flexible_shape(data):
    shape = [
        -1 if len(np.unique(column)) != 1 else column[0] for column in np.asarray([
            sample.shape for sample in data
        ]).transpose()
    ]
    return shape



# Copyright (C) 2024 Matthias Nadig

import torch.nn as nn

from ._base import Module


class SequentialBase(Module):
    def __init__(self):
        super().__init__()

    def create(self, modules, debug):
        self.set_debug(debug=debug)
        self._create(modules=modules)
        return self

    def _create(self, modules):
        """ Takes a list (simple list, not nn.ModuleList!) of tacotree Modules """
        raise NotImplementedError

    def _forward(self, *args):
        raise NotImplementedError


class SequentialSingleInput(SequentialBase):
    """ Wraps nn.Sequential """

    def __init__(self):
        super().__init__()
        self._processor = None

    def _create(self, modules):
        self._processor = nn.Sequential(*modules)

    def _forward(self, *args):
        return self._processor(*args)


class SequentialMultipleInputs(SequentialBase):
    """ Similar to nn.Sequential, but allows first module to take multiple arguments """

    def __init__(self):
        super().__init__()
        self._processors = None

    def _create(self, modules):
        self._processors = nn.ModuleList(modules)

    def _forward(self, *args):
        x = self._processors[0](*args)
        for module in self._processors[1:]:
            x = module(x)
        return x



# Copyright (C) 2024 Matthias Nadig

import numpy as np

import torch.nn as nn

from .. import Batch
from ._modules import Input, Output


class Model(nn.Module):
    def __init__(self, config, sequences, debug=False):
        super().__init__()
        self._config = config
        self._debug = debug

        self._seq = sequences
        flows = self._config.workflow
        assert isinstance(self._seq, nn.ModuleDict) and isinstance(flows, list)

        n_flows = len(flows)
        self._active_flows = np.zeros(n_flows, dtype=int)
        self._is_any_flow_explicitly_activated = False

        # Check if inputs are available for all nodes of the model
        outputs_available = []
        for flow in flows:
            for i, seq_name in enumerate(flow):
                if i == 0:
                    # First sequence of current workflow, check if all inputs are available
                    if isinstance(seq_name, str):
                        seq_name = [seq_name]
                    for input_name in seq_name:
                        is_input_given_or_processed = isinstance(self._seq[input_name], Input) or input_name in outputs_available
                        if not is_input_given_or_processed:
                            raise RuntimeError(
                                f'Given workflow does not allow to process "{input_name}" before required.\n\n' +
                                f'Outputs available:\n{outputs_available}\n\n' +
                                f'Workflow:\n{flows}'
                            )
                else:
                    # This sequence is not the first of current workflow, thus the sequence will have its inputs
                    # provided by the previous sequence of the current workflow
                    if seq_name not in outputs_available:
                        # First time, the current sequence produces an output in any of the workflows
                        outputs_available.append(seq_name)

        self._input_names = []
        self._output_names = []
        for seq_name, seq in self._seq.items():
            if isinstance(seq, Input):
                self._input_names.append(seq_name)
            elif isinstance(seq, Output):
                self._output_names.append(seq_name)
    
    def get_config(self):
        """ Returns config object of this model """
        return self._config

    def get_sequence_names(self):
        """ Returns list of sequence names """
        return list(self._seq.keys())

    def get_sequence(self, name):
        """ Returns sequence by given name """
        return self._seq[name]

    def get_input_names(self):
        """ Returns input names """
        return self._input_names

    def get_output_names(self):
        """ Returns output names """
        return self._output_names

    def clone(self, which=None):
        """ Clones a subgraph of the model (per default the full model) """
        from ._creation_api import create_from_config, create_model
        if which is None:
            # Clone full model
            # 1st: Use config to allocate new tensors for weights
            # 2nd: Copy state dict
            model_cloned = create_from_config(config=self.get_config(), debug=self._debug)
            model_cloned.load_state_dict(self.state_dict())
        else:
            # Clone subgraph of model, check which groups are to be cloned
            if isinstance(which, str):
                groups_clone = [which]
            elif hasattr(which, '__iter__') and all([isinstance(group, str) for group in which]):
                groups_clone = list(which)
            else:
                str_e = \
                    f'Expected string or iterable of strings representing group names, ' + \
                    f'got {type(which)} (value: {which})'
                raise TypeError(str_e)

            # Check if all specified groups exist
            groups_available = self._config.groups
            if not all([group in groups_available for group in groups_clone]):
                str_e = \
                    f'At least one requested group not in group names --> ' + \
                    f'requested {groups_clone}, found {list(groups_available.keys())}'
                raise ValueError(str_e)

            # Get groups, workflows and sequences to be cloned
            groups_clone = {
                group: indices_seq for group, indices_seq in groups_available.items()
                if group in groups_clone
            }
            workflow_clone = [
                sequences for i, sequences in enumerate(self._config.workflow)
                if any([i in indices_seq for _, indices_seq in groups_clone.items()])
            ]
            sequences_clone = {
                name_seq: seq for name_seq, seq in self._config.sequences.items()
                if any([
                    (
                        (name_seq == sequences[0] if isinstance(sequences[0], str) else name_seq in sequences[0]) or
                        name_seq in sequences[1:]
                    ) for i, sequences in enumerate(workflow_clone)
                ])
            }

            # Create new model and copy state dict of all required sequences
            model_cloned = create_model(
                sequences=sequences_clone, workflow=workflow_clone,
                groups=groups_clone, debug=self._debug)
            for sequence_name in model_cloned.get_sequence_names():
                sequence_new = model_cloned.get_sequence(sequence_name)
                sequences_old = self.get_sequence(sequence_name)
                sequence_new.load_state_dict(sequences_old.state_dict())

        return model_cloned

    def active_workflows(self, which):
        """
        Set active workflows explicitly
        If this method is not called, all workflows will be active per default.
        This method will deactivate the ones not explicitly activated here.
        Use in a with-statement, e.g.:
        with model.active_workflows([0, 1, 5, 6]):
            outputs = model(...)
        """
        if isinstance(which, str):
            if which == 'all':
                indices = np.arange(len(self._config.workflow))
            else:
                if which not in self._config.groups:
                    str_e = f'Group name "{which}" not available: {self._config.groups.keys()}'
                    raise ValueError(str_e)
                indices = self._config.groups[which]
        elif hasattr(which, '__iter__') and all([np.issubdtype(type(index), np.integer) for index in which]):
            indices = which
        else:
            str_e = f'Argument bad, should be indices or group name, got: type = {type(which)}, value = {which}'
            raise ValueError(str_e)
        indices = np.asarray(indices, dtype=int)
        return _WorkflowSwitch(self._activate_workflows, self._deactivate_workflows, indices)

    def forward(self, **inputs):
        """
        Forward pass of PyTorch modules
        A bit unusual as inputs should be given as keyword arguments.
        """
        # All inputs should be of batch type
        if any([not isinstance(value, Batch) for _, value in inputs.items()]):
            raise TypeError('All inputs must be of batch type, got {}'.format([
                type(value) for _, value in inputs.items()
            ]))

        # Dictionary that will contain model outputs
        outputs = dict()

        # Process all model branches in right order
        outputs_each_node = dict()
        flows = self._config.workflow
        for index_flow, flow in enumerate(flows):
            # Check if current workflow is deactivated
            if not self._is_workflow_active(index_flow):
                continue

            # Process all nodes of current workflow
            for i in range(1, len(flow)):
                seq_name_inputs = flow[i-1]
                seq_name_process = flow[i]

                if isinstance(seq_name_inputs, str):
                    seq_name_inputs = [seq_name_inputs]

                # Get inputs for current sequence either from this function's arguments or from other sequence's output
                inputs_sequence = []
                for input_name in seq_name_inputs:
                    if isinstance(self._seq[input_name], Input):
                        # TODO: Make sure that input sequence is not run redundantly
                        input_retrieved = self._seq[input_name](inputs[input_name])
                        input_current = input_retrieved
                    else:
                        input_current = outputs_each_node[input_name]
                    inputs_sequence.append(input_current)

                # Get the next sequence/model branch
                sequence = self._seq[seq_name_process]

                # Set a breakpoint, if debugging is enabled
                # Note:
                # Debugging can be enabled when creating or loading a model. This breakpoint does not require
                # a debugger to be specifically run (e.g. as in PyCharm) and will enter the debug mode.
                if self._debug:
                    breakpoint()

                # Run the current sequence/model branch
                output_current_node = sequence(*inputs_sequence)

                # Store transient output of each node of the model
                outputs_each_node[seq_name_process] = output_current_node

                # Store the node's output in final outputs of the model, in case it ends with an output layer
                if isinstance(sequence, Output):
                    outputs.update({seq_name_process: output_current_node})

        return outputs

    def _activate_workflows(self, indices):
        self._active_flows[indices] += 1
        self._check_workflow_activation()

    def _deactivate_workflows(self, indices):
        self._active_flows[indices] -= 1
        self._check_workflow_activation()

    def _check_workflow_activation(self):
        if (self._active_flows != 0).any():
            self._is_any_flow_explicitly_activated = True

    def _is_workflow_active(self, index_flow):
        # Check if any workflows have explicitly been activated and if so, check if requested flow is deactivated
        if self._is_any_flow_explicitly_activated and not self._active_flows[index_flow]:
            is_active = False
        else:
            is_active = True
        return is_active


class _WorkflowSwitch:
    def __init__(self, fn_activate, fn_deactivate, indices):
        self._fn_activate = fn_activate
        self._fn_deactivate = fn_deactivate
        self._indices = indices

        self._fn_activate(self._indices)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._fn_deactivate(self._indices)



# Copyright (C) 2024 Matthias Nadig

from ._base import Module
from ._sequential import SequentialBase
from ._sequential import SequentialMultipleInputs, SequentialSingleInput
from ._m_in_out import Input, Output
from ._m_axis_mod import Flatten, Vectorize, Reshape, View, Transpose, SwapAxes, Concat, Stack, Squeeze, Unsqueeze, Argmax
from ._m_math import Add
from ._m_global_pooling import GlobalPooling, GlobalAverage, GlobalMax
from ._m_padding import CircularPad1d, CircularPad
from ._m_misc import DebugBreakpoint, PassThrough, Lambda, SelectFromTuple, TTModel



# Copyright (C) 2024 Matthias Nadig

import numpy as np
import torch

from ._base_data import BaseData


def as_array(data, annotations=None):
    data = np.asarray(data)
    data = torch.from_numpy(data)
    return Array(data=data, annotations=annotations)


class Array(BaseData):
    def __init__(self, data, annotations):
        super().__init__(data=data, annotations=annotations)

    def shape(self):
        return self._data.shape



# Copyright (C) 2024 Matthias Nadig

from ._base import Module


class DebugBreakpoint(Module):
    """ Sets breakpoints for debugging """

    def __init__(self):
        super().__init__()

    def _forward(self, *args):
        breakpoint()
        return None if len(args) == 0 else (args[0] if len(args) == 1 else args)


class PassThrough(Module):
    """ A layer that does nothing (e.g. for setting breakpoints when debugging) """

    def __init__(self):
        super().__init__()

    def _forward(self, *args):
        return None if len(args) == 0 else (args[0] if len(args) == 1 else args)


class Lambda(Module):
    """ Invokes callable in forward pass """

    def __init__(self, fn):
        super().__init__()
        self._fn = fn

    def _forward(self, *args):
        return self._fn(*args)


class SelectFromTuple(Module):
    """ E.g. after LSTM layer, which has multiple outputs """

    def __init__(self, index):
        super().__init__()
        # self._index = nn.Parameter(torch.tensor(index), requires_grad=False)
        self._index = index

    def _forward(self, x_tuple):
        return x_tuple[self._index]


class TTModel(Module):
    """ A module that executes a tacotree model """

    def __init__(self, order_inputs, order_outputs, config, state_dict):
        super().__init__()
        # Import the creation function here to not run into recursive import
        from .. import create_from_config

        self._order_inputs = order_inputs
        self._order_outputs = order_outputs
        self._model = create_from_config(config=config)

        self._model.load_state_dict(state_dict)

        # self._workflow_switch = self._model.active_workflows(indices=)

    def _forward(self, *inputs):
        dict_inputs = {input_name: input_value for input_name, input_value in zip(self._order_inputs, inputs)}
        outputs = self._model(**dict_inputs)
        if len(self._order_outputs) == 1:
            ret_outputs = outputs[self._order_outputs[0]]
        else:
            ret_outputs = tuple(outputs[output_name] for output_name in self._order_outputs)
        return ret_outputs



# Copyright (C) 2024 Matthias Nadig

from ... import BatchSingleTensor, BatchFlexShape
from ._base import Module


class Input(Module):
    def __init__(self, batch_size, shape):
        super().__init__()
        if batch_size not in [None, -1] and batch_size <= 0:
            raise RuntimeError(f'batch_size must be > 0, got {batch_size}')

        batch_size = -1 if batch_size is None else batch_size
        shape = tuple([-1 if dim is None else dim for dim in shape])

        self._batch_size = batch_size
        self._shape = tuple(shape)

    def _forward(self, x):
        return self._check_input(x)

    def _check_input(self, x):
        n_samples = x.n_samples()
        shape_sample = x.sample_shape()
        if isinstance(x, BatchSingleTensor):
            pass
        elif isinstance(x, BatchFlexShape):
            if all([item != -1 for item in self._shape]):
                raise RuntimeError(f'Flex batch not allowed with fixed shape')
        else:
            raise TypeError(x)
        shape_x = (n_samples, *shape_sample)
        bad_dim = len(shape_sample) != len(self._shape)
        bad_batch_size = self._batch_size != -1 and n_samples != self._batch_size
        bad_grid_shape = any([
            (dim_target != -1 and dim_x != dim_target) for dim_x, dim_target in zip(shape_sample, self._shape)
        ])
        if bad_dim or bad_batch_size or bad_grid_shape:
            str_error = (
                f'Shape of input {tuple(shape_x)} does not match the requirement ' +
                f'{(None if self._batch_size == -1 else self._batch_size,) + self._shape}'
            )
            raise ValueError(str_error)
        return x


class Output(Module):
    def __init__(self):
        super().__init__()

    def _forward(self, *args):
        return None if len(args) == 0 else (args[0] if len(args) == 1 else args)



# Copyright (C) 2024 Matthias Nadig

from ._sample import as_sample, as_sample_with_channels
from ._sample import Sample, SampleChannels
from ._batch import as_batch, as_batch_with_channels
from ._batch import Batch
from ._batch import BatchSingleTensor, BatchFlexShape
from ._batch import BatchChannelMatrix, BatchChannelsSingleTensor, BatchChannelsFlexGrid
from ._array import as_array, Array

from ._annotations import Annotation, raise_if_not_none_or_iterable_of_annotations
from ._annotations import as_label, Label
from ._annotations import as_bounds, Bounds
from ._annotations import as_semantic_map_overlapping, as_semantic_labelmap
from ._annotations import as_instance_map_overlapping, as_instance_labelmap
from ._annotations import SegmentationMap, SegmentationMapMultiClass
from ._annotations import SemanticMapOverlapping, SemanticLabelmap
from ._annotations import InstanceMapOverlapping, InstanceLabelmap



# Copyright (C) 2024 Matthias Nadig

from ._pytorch_layer import PyTorchLayer, safe_instantiation_of_layer
from . import templates
from .templates import Input, Output
from .templates import SequentialSingleInput, SequentialMultipleInputs


_MODULE_CONSTRUCTORS = {
    # Shape
    'Flatten': templates.Flatten,
    'Vectorize': templates.Vectorize,
    'Reshape': templates.Reshape,
    'View': templates.View,
    'Transpose': templates.Transpose,
    'SwapAxes': templates.SwapAxes,
    'Concat': templates.Concat,
    'Stack': templates.Stack,
    'Squeeze': templates.Squeeze,
    'Unsqueeze': templates.Unsqueeze,

    # Math
    'Add': templates.Add,

    # Global pooling
    'GlobalAverage': templates.GlobalAverage,
    'GlobalMax': templates.GlobalMax,

    # Padding
    'CircularPad1d': templates.CircularPad1d,
    'CircularPad': templates.CircularPad,

    # Others
    'Argmax': templates.Argmax,
    'DebugBreakpoint': templates.DebugBreakpoint,
    'PassThrough': templates.PassThrough,
    'Lambda': templates.Lambda,
    'SelectFromTuple': templates.SelectFromTuple,
    'TTModel': templates.TTModel,
}

# Special sequential modules
# Dictionary that provides the constructor for a sequential for some modules.
# Such constructors can be given by the user if a set of custom modules is more
# efficient with a special way of handling sequential execution.
# Example: Custom pipeline class that should be used with some custom modules for
# preprocessing or augmentation.
_MODULES_WITH_SPECIAL_SEQUENTIALS = {}


def create_sequence(modules, debug, has_multiple_inputs):
    if len(modules) == 1:
        # Sequence of a single module (does not require a sequential wrapper)
        module_info = modules[0]
        # (name, params) = module_info  # params can sometimes not be specified
        seq = _get_module(*module_info, debug=debug)
    else:
        # Sequence has multiple modules, choose the appropriate sequential wrapper
        module_names = [m[0] for m in modules]
        constructor_sequential = _get_sequential_constructor(module_names=module_names,
                                                             has_multiple_inputs=has_multiple_inputs)
        modules_built = [_get_module(*module_info, debug=debug) for module_info in modules]
        seq = constructor_sequential().create(modules=modules_built, debug=debug)

    return seq


def add_module(name, constructor, sequential_constructor=None):
    # Use shorter names for variables, otherwise code becomes too much text
    c = constructor
    sc = sequential_constructor
    del constructor, sequential_constructor

    # Check input types
    if not isinstance(name, str):
        str_e = f'Name must be a string, got {type(name)}'
        raise TypeError(str_e)
    elif name in _MODULE_CONSTRUCTORS:
        str_e = f'A module type with name "{name}" already exists'
        raise ValueError(str_e)
    if not issubclass(c, templates.Module):
        str_e = f'Constructor must be of Module type, got {type(c)}'
        raise TypeError(str_e)
    if sc is not None:
        is_valid_sequential = \
            issubclass(sc, templates.SequentialBase) or \
            (hasattr(sc, '__iter__') and all([issubclass(item, templates.SequentialBase) for item in sc]))
        if not is_valid_sequential:
            str_e = \
                f'Sequential constructor must be None, SequentialBase or iterable of SequentialBase, got {type(sc)}'
            raise TypeError(str_e)
        elif name in _MODULES_WITH_SPECIAL_SEQUENTIALS:
            str_e = f'Sequential constructor for module already exists'
            raise RuntimeError(str_e)

    # Add module
    _MODULE_CONSTRUCTORS[name] = c

    # Add special sequential constructor, if given
    if sc is not None:
        _MODULES_WITH_SPECIAL_SEQUENTIALS[name] = sc


def _get_module(name, params=None, debug=False):
    if name == 'Input':
        module_class = Input
    elif name == 'Output':
        module_class = Output
    elif name in _MODULE_CONSTRUCTORS:
        module_class = _MODULE_CONSTRUCTORS[name]
    else:
        # Note:
        # If a module name is not available, this else-statement will be entered but the error
        # will be thrown below in module_class(**params).
        module_class = _get_pytorch_wrapper(name=name)

    params = dict() if params is None else params

    try:
        module = module_class(**params)
        module = module.set_debug(debug=debug)
    except AttributeError:  # as e:
        names_available = list(_MODULE_CONSTRUCTORS.keys()) + ['PyTorch modules in torch.nn']
        n_available = len(names_available)
        str_error = (
            f'Module not found\n\nAvailable modules:' + ('\n\t* {}' * n_available).format(*names_available) +
            f'\n\nModule "{name}" not found. For a list of available modules, see above.'
        )
        raise ValueError(str_error)

    return module


def _get_sequential_constructor(module_names, has_multiple_inputs):
    # Get special constructors for each module
    special_constructors = [
        _MODULES_WITH_SPECIAL_SEQUENTIALS[name] if name in _MODULES_WITH_SPECIAL_SEQUENTIALS else None
        for name in module_names
    ]
    special_constructors_found = []
    for special_constructor in special_constructors:
        if not hasattr(special_constructor, '__iter__'):
            special_constructor = [special_constructor]
        for special_constructor_current in special_constructor:
            if special_constructor_current not in special_constructors_found:
                special_constructors_found.append(special_constructor_current)

    # Check if all modules are eligible for one and the same special constructor
    special_constructors_valid = []
    for special_constructor in special_constructors_found:
        if all([
            special_constructor == c_check or
            (hasattr(c_check, '__iter__') and special_constructor in c_check)
            for c_check in special_constructors
        ]):
            special_constructors_valid.append(special_constructor)

    # Pick the first one (this will be None, if no special constructors are registered)
    if len(special_constructors_valid) > 0:
        special_constructor_used = special_constructors_valid[0]
    else:
        special_constructor_used = None

    # Choose constructor for sequential modules
    if special_constructor_used is None:
        # Use the default one
        if has_multiple_inputs:
            constructor_sequential = SequentialMultipleInputs
        else:
            constructor_sequential = SequentialSingleInput
    else:
        # Use the chosen special constructor
        constructor_sequential = special_constructor_used

    return constructor_sequential


def _get_pytorch_wrapper(name):
    """ Returns a callable that initializes a PyTorch layer """
    def wrap_pytorch_layer(**params):
        return PyTorchLayer(name=name, **params)
    return wrap_pytorch_layer



# Copyright (C) 2024 Matthias Nadig

# DO NOT REMOVE!!!
# Importing torch is required for eval function!
import torch
# DO NOT REMOVE!!!

from ._base import Module


class PyTorchLayer(Module):
    """
    Helper to extract data from batch type, when using inbuilt torch modules
    """

    def __init__(self, name, **params):
        super().__init__()
        layer_class = safe_instantiation_of_layer(name_layer=name)
        self._processor = layer_class(**params)

    def _forward(self, *args):
        n_inputs = len(args)
        if n_inputs == 1:
            # Unpack the batch
            batch = args[0]
            x, annotations, limits = batch.data(), batch.annotations(), batch.limits()

            # Run PyTorch module
            x = self._processor(x)

            # Caveat!
            # Bounding boxes and similar annotations might be inaccurate after this layer in the current implementation.
            # This is due to PyTorch modules that change the shape of the data. For now, we therefore do not
            # forward the annotations here.
            out = self._as_batch(data=x, annotations=None, limits=None)
        else:
            str_e = f'Currently only single inputs are allowed to be given to PyTorch modules'
            raise ValueError(str_e)

        return out


def safe_instantiation_of_layer(name_layer):
    str_reference = f'torch.nn.{name_layer}'
    return _safe_instantiation(str_reference)


def _safe_instantiation(str_reference):
    """
    CAVEAT!
    This function uses eval(). Therefore, it first checks the input for alphanumeric
    characters to ensure, no malicious code is executed.
    """

    # Throw if module name
    #   - contains non-alphanumeric characters (except for "." and "_") or
    #   - does not start with "torch."
    chars_check = str_reference.replace('.', '').replace('_', '')
    is_alphanumeric = chars_check.isascii() and chars_check.isalnum()
    is_torch_attribute = str_reference.startswith('torch.')
    if not is_alphanumeric or not is_torch_attribute:
        str_e = f'Got input that is non-alphanumeric or does not reference an attribute of torch: "{str_reference}"'
        raise ValueError(str_e)

    # Get the required attribute
    attr = eval(str_reference)

    return attr



# Copyright (C) 2024 Matthias Nadig

import numpy as np
import ndbounds


def as_label(label):
    return Label(label=label)


def as_bounds(bounds, labels):
    return Bounds(bounds=bounds, labels=labels)


def as_semantic_map_overlapping(segmentation_map, labels, n_dims):
    _check_input_types(segmentation_map=segmentation_map, labels=labels, dtype_required=np.bool_)
    _check_grid_shape(segmentation_map=segmentation_map, n_dims=n_dims, n_leading=1)
    return SemanticMapOverlapping(segmentation_map=segmentation_map, labels=labels)


def as_semantic_labelmap(segmentation_map, labels, n_dims):
    _check_input_types(segmentation_map=segmentation_map, labels=labels, dtype_required=np.integer)
    _check_grid_shape(segmentation_map=segmentation_map, n_dims=n_dims, n_leading=0)
    return SemanticLabelmap(segmentation_map=segmentation_map, labels=labels)


def as_instance_map_overlapping(segmentation_map, labels, n_dims):
    _check_input_types(segmentation_map=segmentation_map, labels=labels, dtype_required=np.bool_)
    _check_grid_shape(segmentation_map=segmentation_map, n_dims=n_dims, n_leading=2)
    return InstanceMapOverlapping(segmentation_map=segmentation_map, labels=labels)


def as_instance_labelmap(segmentation_map, labels, n_dims):
    _check_input_types(segmentation_map=segmentation_map, labels=labels, dtype_required=np.integer)
    _check_grid_shape(segmentation_map=segmentation_map, n_dims=n_dims, n_leading=1)
    return InstanceLabelmap(segmentation_map=segmentation_map, labels=labels)


def _check_input_types(segmentation_map, labels, dtype_required):
    if not isinstance(segmentation_map, np.ndarray) or not np.issubdtype(segmentation_map.dtype, dtype_required):
        str_e = f'Expected segmentation_map to be array of type {dtype_required}, got {type(segmentation_map)}'
        if isinstance(segmentation_map, np.ndarray) and not np.issubdtype(segmentation_map.dtype, dtype_required):
            str_e += f' of type {segmentation_map.dtype}'
        raise TypeError(str_e)
    if not isinstance(labels, np.ndarray):
        str_e = f'Expected labels to be array, got {type(labels)}'
        raise TypeError(str_e)


def _check_grid_shape(segmentation_map, n_dims, n_leading):
    if segmentation_map.ndim - n_leading != n_dims:
        str_e = \
            f'Bad shape {segmentation_map.shape} for this kind of segmentation map assuming a {n_dims}-dimensional grid'
        raise ValueError(str_e)


def raise_if_not_none_or_iterable_of_annotations(annotations):
    if annotations is not None and (not hasattr(annotations, '__iter__') or
                                    not all([isinstance(ann, Annotation) for ann in annotations])):
        raise TypeError(f'Annotations should be a list of Annotation-objects, got {annotations}')


class Annotation:
    """
    Base class for annotations
    """
    def __init__(self):
        pass


class Label(Annotation):
    """
    Label (could be of any type: integer, string, ...)
    """
    def __init__(self, label):
        super().__init__()
        self._label = label

    def label(self):
        return self._label


class Bounds(Annotation):
    """
    Bounds (N-dimensional, start and end of objects) with label
    """
    def __init__(self, bounds, labels):
        super().__init__()
        ndbounds.raise_if_not_ndbounds(bounds=bounds)

        self._bounds = bounds
        self._labels = np.asarray(labels)

    def bounds(self):
        return self._bounds

    def labels(self):
        return self._labels

    def n_objects(self):
        return len(self._labels)


class SegmentationMap(Annotation):
    """
    Base class for various kinds of segmentation maps
    """
    def __init__(self, segmentation_map, labels, n_dims):
        super().__init__()
        assert isinstance(segmentation_map, np.ndarray)
        assert isinstance(labels, np.ndarray)
        self._map = segmentation_map
        self._labels = labels
        self._n_dims = n_dims

    def n_dims(self):
        """ Returns number of dimensions of grid """
        return self._n_dims

    def shape(self):
        """ Returns shape of grid """
        return self._map.shape[-self._n_dims:]


class SegmentationMapMultiClass(SegmentationMap):
    """
    Base class for various kinds of segmentation maps given multiple classes
    """
    def __init__(self, segmentation_map, labels, n_dims):
        super().__init__(segmentation_map=segmentation_map, labels=labels, n_dims=n_dims)

    def n_classes(self):
        return len(self._labels)

    def get_classes(self):
        return self._labels

    def semantic_map(self, label):
        """
        Returns semantic map for a given label
        Shape: (dim 1, ..., dim N)
        """
        return self._semantic_map(label)

    def _index_of_label(self, label):
        return np.argwhere(self._labels == label).flatten()[0]

    def _semantic_map(self, label):
        raise NotImplementedError


class SemanticMapOverlapping(SegmentationMapMultiClass):
    """
    Semantic segmentation map
    Maps for all classes are stacked as boolean masks.
    Shape: (classes, dim 1, ..., dim N)
    """
    def __init__(self, segmentation_map, labels):
        n_dims = segmentation_map.ndim - 1
        super().__init__(segmentation_map=segmentation_map, labels=labels, n_dims=n_dims)

    def _semantic_map(self, label):
        semantic_map = self._map[self._index_of_label(label)]
        return semantic_map


class SemanticLabelmap(SegmentationMapMultiClass):
    """
    Semantic segmentation map
    Single map that encodes classes using integers.
    Shape: (dim 1, ..., dim N)
    """
    def __init__(self, segmentation_map, labels):
        n_dims = segmentation_map.ndim
        super().__init__(segmentation_map=segmentation_map, labels=labels, n_dims=n_dims)

    def _semantic_map(self, label):
        semantic_map = self._map == self._index_of_label(label)
        return semantic_map

    def semantic_labelmap(self):
        return self._map


class InstanceMapOverlapping(SegmentationMapMultiClass):
    """
    Instance segmentation map
    Maps for all classes are stacked as boolean masks.
    Boolean type with shape: (classes, instances, dim 1, ..., dim N)
    """
    def __init__(self, segmentation_map, labels):
        n_dims = segmentation_map.ndim - 2
        super().__init__(segmentation_map=segmentation_map, labels=labels, n_dims=n_dims)
    
    def _semantic_map(self, label):
        semantic_map = self._map[self._index_of_label(label)].any(axis=0)
        return semantic_map


class InstanceLabelmap(SegmentationMapMultiClass):
    """
    Instance segmentation map
    Maps for all classes are stacked and encode instances using integers.
    Integer type with shape: (classes, dim 1, ..., dim N)
    """
    def __init__(self, segmentation_map, labels):
        n_dims = segmentation_map.ndim - 1
        super().__init__(segmentation_map=segmentation_map, labels=labels, n_dims=n_dims)
    
    def _semantic_map(self, label):
        semantic_map = self._map[self._index_of_label(label)].astype(bool)
        return semantic_map



# Copyright (C) 2024 Matthias Nadig

import torch

from ._base import Module


def _axis_in_batch_dim(axis):
    """
    Checks if axis >= 0 and then adds 1, as parameters given by user should not include batch size
    Note:
        All axes should be given relating to a single sample, the batch axis (the 1st one) should be added and taken
        care of by this toolbox. Therefore, negative axes should be accepted only in special cases as in Flatten.
        To allow the usage of more dynamic negative axes, the forward function should check, if the batch axis is
        affected.
    """
    if not isinstance(axis, int) or axis < 0:
        str_e = f'Axis must be an integer >= 0, got {axis} of type {type(axis)}'
        raise RuntimeError(str_e)
    return axis + 1


class Flatten(Module):
    def __init__(self, axis_start=0, axis_end=-1):
        super().__init__()
        # Parameters given by user should not include batch size
        axis_start = _axis_in_batch_dim(axis_start)
        axis_end = axis_end if axis_end == -1 else _axis_in_batch_dim(axis_end)
        self._axis_start = axis_start
        self._axis_end = axis_end

    def _forward(self, x):
        x = x.data()
        x = torch.flatten(x, start_dim=self._axis_start, end_dim=self._axis_end)
        return self._as_batch(data=x, annotations=None, limits=None)


class Vectorize(Flatten):
    def __init__(self):
        super().__init__(axis_start=0)


class Transpose(Module):
    def __init__(self, order):
        super().__init__()
        if not hasattr(order, '__iter__'):
            raise TypeError(f'order must be an iterable, got value of type {type(order)} (value: {order})')

        # Parameters given by user should not include batch size
        order = (0,) + tuple([_axis_in_batch_dim(axis) for axis in order])
        self._order = order

    def _forward(self, x):
        x = x.data()
        x = x.permute(*self._order)
        return self._as_batch(data=x, annotations=None, limits=None)


class SwapAxes(Module):
    def __init__(self, axis1, axis2):
        super().__init__()
        # Parameters given by user should not include batch size
        axis1 = _axis_in_batch_dim(axis1)
        axis2 = _axis_in_batch_dim(axis2)
        self._axis1 = axis1
        self._axis2 = axis2

    def _forward(self, x):
        x = x.data()
        x = x.swapaxes(self._axis1, self._axis2)
        return self._as_batch(data=x, annotations=None, limits=None)


class Concat(Module):
    def __init__(self, axis):
        super().__init__()
        # Parameters given by user should not include batch size
        axis = _axis_in_batch_dim(axis)
        self._axis = axis

    def _forward(self, *args):
        args = [arg.data() for arg in args]
        x = torch.cat(args, self._axis)
        return self._as_batch(data=x, annotations=None, limits=None)


class Stack(Module):
    def __init__(self, axis):
        super().__init__()
        # Parameters given by user should not include batch size
        axis = _axis_in_batch_dim(axis)
        self._axis = axis

    def _forward(self, *args):
        args = [arg.data() for arg in args]
        x = torch.stack(args, self._axis)
        return self._as_batch(data=x, annotations=None, limits=None)


class Squeeze(Module):
    def __init__(self, axis):
        super().__init__()
        # Parameters given by user should not include batch size
        axis = _axis_in_batch_dim(axis)
        self._axis = axis

    def _forward(self, x):
        x = x.data()
        x = x.squeeze(self._axis)
        return self._as_batch(data=x, annotations=None, limits=None)


class Unsqueeze(Module):
    def __init__(self, axis):
        super().__init__()
        # Parameters given by user should not include batch size
        axis = _axis_in_batch_dim(axis)
        self._axis = axis

    def _forward(self, x):
        x = x.data()
        x = x.unsqueeze(self._axis)
        return self._as_batch(data=x, annotations=None, limits=None)


class Reshape(Module):
    def __init__(self, shape):
        super().__init__()
        if not hasattr(shape, '__iter__'):
            raise TypeError(f'shape must be an iterable, got value of type {type(shape)} (value: {shape})')
        self._shape = tuple(shape)

    def _forward(self, x):
        x = x.data()
        n_samples = len(x)
        x = x.reshape(n_samples, *self._shape)
        return self._as_batch(data=x, annotations=None, limits=None)


class View(Module):
    def __init__(self, shape):
        super().__init__()
        if not hasattr(shape, '__iter__'):
            raise TypeError(f'shape must be an iterable, got value of type {type(shape)} (value: {shape})')
        self._shape = tuple(shape)

    def _forward(self, x):
        x = x.data()
        n_samples = len(x)
        x = x.view(n_samples, *self._shape)
        return self._as_batch(data=x, annotations=None, limits=None)


class Argmax(Module):
    def __init__(self, axis=0):
        super().__init__()
        # Parameters given by user should not include batch size
        axis = _axis_in_batch_dim(axis)
        self._axis = axis

    def _forward(self, x):
        x = x.data()
        x = torch.argmax(x, self._axis)
        return self._as_batch(data=x, annotations=None, limits=None)



# Copyright (C) 2024 Matthias Nadig

import numpy as np
import torch

from ._base_data import BaseData
from . import _annotations


def as_sample(data, n_dims, annotations=None, limits=None):
    return _as_sample(data=data, n_dims=n_dims, annotations=annotations, limits=limits, constructor=Sample)


def as_sample_with_channels(data, n_dims, annotations=None, limits=None):
    return _as_sample(data=data, n_dims=n_dims, annotations=annotations, limits=limits, constructor=SampleChannels)


def _as_sample(constructor, data, n_dims, annotations=None, limits=None):
    # Make sure inputs are of correct type
    if not isinstance(n_dims, (int, float)) or n_dims != int(n_dims):
        raise TypeError(f'Expected integer for number of dimensions, got {n_dims} (type: {type(n_dims)})')

    if not isinstance(data, torch.Tensor):
        data = np.asarray(data)
        data = torch.from_numpy(data)
        _annotations.raise_if_not_none_or_iterable_of_annotations(annotations)

    # Data should be in the format (channels, dim 1, dim 2, ..., dim n)
    # Note: Shape can also be (,). In this case, n_dim must be -1 to work seamlessly with batch creation.
    if data.ndim != 1 + n_dims:
        raise ValueError(f'Got data with shape {data.shape} for {n_dims}-dimensional data')

    return constructor(data=data, annotations=annotations, limits=limits)


class Sample(BaseData):
    def __init__(self, data, annotations, limits=None):
        super().__init__(data=data, annotations=annotations)
        self._shape = self._data.shape
        self._n_axes = len(self._shape)
        self._limits = None if limits is None else np.array(limits)

    def set_data(self, data):
        """ CAVEAT! Should only be used if modified data has same properties """
        if data.shape != self._data.shape or data.dtype != self._data.dtype:
            raise ValueError(f'Trying to set not matching data (old/new): {self._data.shape} != {data.shape}, {self._data.dtype} != {data.dtype}')
        self._data = data
        return self

    def shape(self):
        return self._shape

    def n_axes(self):
        return self._n_axes

    def limits(self):
        return self._limits

    def set_limits(self, limits):
        self._limits = limits
        return self


class SampleChannels(Sample):
    def __init__(self, data, annotations, limits=None):
        super().__init__(data=data, annotations=annotations, limits=limits)
        self._n_channels = self._data.shape[0]
        self._grid_shape = self._data.shape[1:]
        self._n_dims = len(self._grid_shape)

    def n_channels(self):
        return self._n_channels

    def n_dims(self):
        return self._n_dims

    def grid_shape(self):
        return self._grid_shape



# Copyright (C) 2024 Matthias Nadig

import torch
import torch.nn.functional as F

from ._base import Module


_VECTORIZE = False


class GlobalPooling(Module):
    def __init__(self, fn_pooling, n_dims, n_features_per_channel, vectorize):
        super().__init__()
        self._fn_pooling = fn_pooling
        if isinstance(n_dims, int):
            self._n_dims = n_dims
        else:
            raise TypeError(f'Expected integer, got value of type {type(n_dims)} (value: {n_dims})')
        self.n_features_per_channel = n_features_per_channel
        self._vectorize = vectorize

    def _forward(self, x):
        data = x.data()
        data = self._fn_pooling(data, n_dims=self._n_dims)
        if not self._vectorize:
            data = data.view((-1,)*len(data.shape) + (self.n_features_per_channel,))
        x = self._as_batch_with_channels(data=data, annotations=None, limits=None)
        return x


class GlobalAverage(GlobalPooling):
    def __init__(self, n_dims, vectorize=_VECTORIZE):
        super().__init__(fn_pooling=_get_global_average, n_dims=n_dims, n_features_per_channel=1, vectorize=vectorize)


class GlobalMax(GlobalPooling):
    def __init__(self, n_dims, vectorize=_VECTORIZE):
        super().__init__(fn_pooling=_get_global_maximum, n_dims=n_dims, n_features_per_channel=1, vectorize=vectorize)


def _get_global_average(x, n_dims):
    size_out = (1,) * n_dims

    if n_dims == 1:
        x = F.adaptive_avg_pool1d(x, size_out)
    elif n_dims == 2:
        x = F.adaptive_avg_pool2d(x, size_out)
    elif n_dims == 3:
        x = F.adaptive_avg_pool3d(x, size_out)
    else:
        raise ValueError(n_dims)

    for _ in range(n_dims):
        x = torch.squeeze(x, -1)
    return x


def _get_global_maximum(x, n_dims):
    size_out = (1,) * n_dims

    if n_dims == 1:
        x = F.adaptive_max_pool1d(x, size_out)
    elif n_dims == 2:
        x = F.adaptive_max_pool2d(x, size_out)
    elif n_dims == 3:
        x = F.adaptive_max_pool3d(x, size_out)
    else:
        raise ValueError(n_dims)

    for _ in range(n_dims):
        x = torch.squeeze(x, -1)
    return x



# Copyright (C) 2024 Matthias Nadig

import torch.nn as nn

from ... import as_batch, as_batch_with_channels  # , BatchSingleTensor, BatchFlexGrid


class Module(nn.Module):
    """
    Every module should be of this class (network layers, preprocessing, ...) even if they are stacked in a Sequential
    """

    def __init__(self):
        super().__init__()
        self._debug = False

    def set_debug(self, debug):
        self._debug = debug
        return self

    def forward(self, *batches_in):
        batches_out = self._forward(*batches_in)

        if self._debug:
            breakpoint()

        return batches_out

    @staticmethod
    def _as_batch(data, annotations, limits):
        return as_batch(samples=data, annotations=annotations, limits=limits)

    @staticmethod
    def _as_batch_with_channels(data, annotations, limits):
        return as_batch_with_channels(samples=data, annotations=annotations, limits=limits)

    def _forward(self, *args):
        raise NotImplementedError



# Copyright (C) 2024 Matthias Nadig

import torch.nn as nn

from ._modules import add_module
from . import Config, Model
from . import _modules


def create_model(sequences, workflow, groups=None, debug=False):
    """
    Sequences specify each branch as sequential sub-model.
    Workflow specifies how inputs are passed from branch to branch.
    """
    config = _create_config(sequences=sequences, workflow=workflow, groups=groups)
    model = _create_from_config(config=config, debug=debug)
    return model


def create_config(sequences, workflow, groups=None):
    config = _create_config(sequences=sequences, workflow=workflow, groups=groups)
    return config


def create_from_config(config, debug=False):
    return _create_from_config(config=config, debug=debug)


def _create_config(sequences, workflow, groups):
    config = Config(sequences=sequences, workflow=workflow, groups=groups)
    return config


def _create_from_config(config, debug):
    module_dict = dict()
    for seq_name, seq_modules in config.sequences.items():
        # Check if sequence has multiple inputs at some point of the workflow
        # (as of now, we need to know this for choosing nn.Sequential or not)
        flag_per_flow = []
        for flow in config.workflow:
            for i, seq_names_check in enumerate(flow):
                if i != 0:
                    if isinstance(seq_names_check, str):
                        seq_names_check = [seq_names_check]
                    if seq_name in seq_names_check:
                        seq_names_inputs = flow[i-1]
                        if not isinstance(seq_names_inputs, str) and len(seq_names_inputs) > 1:
                            has_multiple_inputs = True
                            break
            else:
                has_multiple_inputs = False
            flag_per_flow.append(has_multiple_inputs)
        has_multiple_inputs = any(flag_per_flow)

        # Create the sequence
        module_dict[seq_name] = _modules.create_sequence(modules=seq_modules, debug=debug,
                                                         has_multiple_inputs=has_multiple_inputs)
    sequences = nn.ModuleDict(module_dict)

    model = Model(config=config, sequences=sequences, debug=debug)

    return model



# Copyright (C) 2024 Matthias Nadig

import os
import json

import torch

from ._creation import create_from_config
from ._creation import Config
from ._creation import Model


_FILETYPE_MODEL = 'ttmodel'
_FILETYPE_CONFIG = 'ttconfig'
_FILETYPE_CHECKPOINT = 'ttcheckpoint'


def to_storable_dict(model):
    obj = dict(
        config=model.get_config().to_json_compatible(),
        state_dict=model.state_dict(),
    )
    return obj


def from_storable_dict(obj, debug=False):
    config = Config.from_json_compatible(obj['config'])
    state_dict = obj['state_dict']
    model = create_from_config(config=config, debug=debug)
    model.load_state_dict(state_dict)
    return model


def save_model(filename, model, overwrite=False, makedirs=False):
    filename = _add_appendix_model(filename)

    if os.path.isfile(filename) and not overwrite:
        raise FileExistsError(f'"{filename}"')
    if makedirs:
        dirname = os.path.dirname(filename)
        os.makedirs(dirname, exist_ok=True)

    obj = to_storable_dict(model)
    torch.save(obj, filename)


def load_model(filename, debug=False):
    filename = _add_appendix_model(filename)

    map_location = torch.device('cpu')
    obj = torch.load(filename, map_location=map_location)
    model = from_storable_dict(obj, debug=debug)

    return model


def save_config(filename, obj, overwrite=False, makedirs=False):
    filename = _add_appendix_config(filename)

    if os.path.isfile(filename) and not overwrite:
        raise FileExistsError(f'"{filename}"')
    if makedirs:
        dirname = os.path.dirname(filename)
        os.makedirs(dirname, exist_ok=True)

    if isinstance(obj, Model):
        # Read config of given model
        obj = obj.config.to_json_compatible()
    elif isinstance(obj, Config):
        # Config is given directly
        obj = obj.to_json_compatible()
    else:
        raise TypeError(f'Expected object of type Model or Config, got {type(obj)}')

    with open(filename, 'w') as f:
        json.dump(obj, f)


def load_config(filename):
    filename = _add_appendix_config(filename)

    with open(filename, 'r') as f:
        config = json.load(f)
    config = Config.from_json_compatible(config)

    return config


def save_checkpoint(filename, model, metadata, overwrite=False, makedirs=False):
    filename = _add_appendix_checkpoint(filename)

    if os.path.isfile(filename) and not overwrite:
        raise FileExistsError(f'"{filename}"')
    if makedirs:
        dirname = os.path.dirname(filename)
        os.makedirs(dirname, exist_ok=True)

    obj = dict(
        model=to_storable_dict(model),
        metadata=metadata,
    )
    torch.save(obj, filename)


def load_checkpoint(filename, debug=False):
    filename = _add_appendix_checkpoint(filename)

    map_location = torch.device('cpu')
    obj = torch.load(filename, map_location=map_location)
    model = from_storable_dict(obj['model'], debug=debug)
    metadata = obj['metadata']

    return model, metadata


def _add_appendix_model(filename):
    return _add_appendix(filename, _FILETYPE_MODEL)


def _add_appendix_config(filename):
    return _add_appendix(filename, _FILETYPE_CONFIG)


def _add_appendix_checkpoint(filename):
    return _add_appendix(filename, _FILETYPE_CHECKPOINT)


def _add_appendix(filename, filetype):
    appendix = '.' + filetype
    return filename + appendix if not filename.endswith(appendix) else filename



# Copyright (C) 2024 Matthias Nadig

from setuptools import setup, find_packages
import warnings


with open('README.md', 'r') as f:
    long_description = f.read()

path_package_toplevel = 'src'

# Minimal Python version
# (3.7 is required for breakpoint function)
python_requires = '>=3.7'

# Required PyTorch packages
pytorch_packages = [
    'torch',
    'torchvision',
]

# PyTorch should be installed separately, as there are sometimes issues with the installation
install_requires = [
    'tacotree_lab',
    'numpy',
    'ndbounds',
    *pytorch_packages,
]

# Check if PyTorch has already been installed separately
missing_packages = []
for package in pytorch_packages:
    # Make sure that package name contains nothing but ASCII and alphanumeric characters
    if not all([char.isascii() and char.isalnum() for char in package]):
        str_e = f'Package name has unexpected characters: "{package}"'
        raise RuntimeError(str_e)

    # Import the package test-wise to see if it is there
    try:
        exec(f'import {package}')
    except ImportError:
        missing_packages.append(package)

# Throw error if any PyTorch package is not available
if len(missing_packages) > 0:
    str_e = (
        'At least one PyTorch package is missing. ' +
        'PyTorch should be installed separately from tacotree, as its installation ' +
        'can sometimes be tricky and hardware-dependent. ' +
        'Following packages are missing:' +
        ('\n\t* {}' * len(missing_packages)).format(*missing_packages)
    )
    warnings.warn(str_e)

setup(
    name='tacotree',
    version='1.0.0',
    description='Data processing pipelines based on PyTorch',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Matthias Nadig',
    author_email='nadig_develop@yahoo.com',
    license='MIT',
    package_dir={'': path_package_toplevel},
    packages=find_packages(where=path_package_toplevel),
    python_requires=python_requires,
    install_requires=install_requires,
)



# Copyright (C) 2024 Matthias Nadig

import numpy as np

import torch

from ._base import Module


class CircularPad(Module):
    def __init__(self, n_dims, pad):
        super().__init__()
        self._n_dims = n_dims
        if isinstance(pad, int):
            pad = (pad,) * self._n_dims * 2
        if not (
            hasattr(pad, '__iter__') and
            len(pad) == self._n_dims * 2 and
            all([isinstance(p, int) for p in pad])
        ):
            raise TypeError(f'Padding info invalid: {pad}')
        self._pad = np.array(pad).reshape((self._n_dims, 2))

    def _forward(self, batch):
        x, annotations, limits = batch.data(), batch.annotations(), batch.limits()

        n_axes_input = x.ndim
        n_dims_pooling = self._n_dims
        n_axes_static = n_axes_input - n_dims_pooling

        shape_dims = x.shape[-n_dims_pooling:]

        slice_full = slice(0, None, 1)
        slices_static = (slice_full,) * n_axes_static

        for i in range(n_dims_pooling):
            slice_pad_right = slices_static + tuple(
                slice_full if index_dim != i else slice(0, self._pad[i, 1], 1)
                for index_dim in range(n_dims_pooling)
            )
            slice_pad_left = slices_static + tuple(
                slice_full if index_dim != i else slice(shape_dims[i]-self._pad[i, 0], shape_dims[i], 1)
                for index_dim in range(n_dims_pooling)
            )

            x = torch.cat([x, x[slice_pad_right]], dim=n_axes_static+i)
            x = torch.cat([x[slice_pad_left], x], dim=n_axes_static+i)

        # Caveat!
        # Bounding boxes and similar annotations are inaccurate after this layer in the current implementation.
        # For now, we therefore do not forward the annotations here.
        x = self._as_batch(data=x, annotations=None, limits=None)

        return x


class CircularPad1d(CircularPad):
    def __init__(self, pad):
        super().__init__(n_dims=1, pad=pad)


if __name__ == '__main__':
    a = np.array([
        [1, 2, 3, 4, 5],
        [11, 22, 33, 44, 55],
        [111, 222, 333, 444, 555],
        [1111, 2222, 3333, 4444, 555],
    ])
    t = torch.from_numpy(np.repeat(a[np.newaxis], 2, axis=0))

    print(f'Original: {t}')

    t_pad = CircularPad1d(pad=(2, 3))(t)

    print(f'Padded: {t_pad}')



# Copyright (C) 2024 Matthias Nadig


class BaseData:
    def __init__(self, data, annotations):
        self._data = data
        self._ann = annotations

    def data(self):
        return self._data

    def annotations(self):
        return self._ann

    def as_type(self, dtype):
        self._data = self._data.type(dtype)
        return self

    def to(self, *args, **kwargs):
        self._data = self._data.to(*args, **kwargs)
        return self

    def enable_grad(self, requires_grad):
        self._data.requires_grad_(requires_grad)
        return self



# Copyright (C) 2024 Matthias Nadig

import io
import base64
import numpy as np

from ._model import Model


class Config:
    # Version for compatibility within this toolbox
    VERSION_MAX = 4

    def __init__(self, sequences, workflow, groups=None, version=None):
        super().__init__()
        self.version = self.VERSION_MAX if version is None else version
        self.workflow = workflow

        # self.groups = {'all': [i for i in range(len(self.workflow))]}
        if groups is not None:
            if 'all' in groups:
                str_e = f'Workflow groups have been specified by user and contain invalid group name "all": {groups}'
                raise ValueError(str_e)
            # self.groups = dict(**self.groups, **groups)
            self.groups = groups
        else:
            self.groups = dict()

        sequences_set = dict()
        for seq_name, seq in sequences.items():
            sequences_set[seq_name] = []
            for module in seq:
                if isinstance(module, Model):
                    module = (
                        'TTModel',
                        dict(
                            order_inputs=module.get_input_names(),
                            order_outputs=module.get_output_names(),
                            config=module.get_config(),
                            state_dict=module.state_dict(),
                        )
                    )
                elif module[0] == 'Source':
                    # How to work with Array data? --> axis modification etc. currently designed to work on samples
                    raise NotImplementedError('Implement Source module in model first')
                    # Make sure that data of source is JSON-serializable in config
                    data = module[1]['data']
                    data = np.asarray(data)
                    dtype_source = data.dtype
                    # if not np.issubdtype(dtype_source, np.number):  # this would allow complex numbers
                    if not (np.issubdtype(dtype_source, np.integer) or np.issubdtype(dtype_source, np.floating)):
                        str_e = f'Data type {dtype_source} not allowed for Source module'
                        raise ValueError(str_e)
                    with io.BytesIO() as sink:
                        np.save(sink, data)
                        bytes_data = sink.getvalue()
                    str_data = base64.b64encode(bytes_data).decode()
                    module = (
                        module[0],
                        dict(data=str_data)
                    )
                else:
                    pass
                sequences_set[seq_name].append(module)
        self.sequences = sequences_set

    @staticmethod
    def from_json_compatible(obj):
        # Ignore version 1 and 2
        version = 1 if 'version' not in obj else obj['version']
        if version == 3:
            if 'groups' in obj:
                str_e = f'Groups not expected to be given for model of version {version} (obj = {obj})'
                raise RuntimeError(str_e)
            obj['groups'] = None
        else:
            if version != Config.VERSION_MAX:
                raise RuntimeError(f'Version {version} not compatible (accepted versions: {Config.VERSION_MAX})')
        return Config(sequences=obj['sequences'], workflow=obj['workflow'], groups=obj['groups'], version=version)

    def to_json_compatible(self):
        obj = dict(
            version=self.version,
            sequences=self.sequences,
            workflow=self.workflow,
            groups=self.groups,
        )
        return obj



# Copyright (C) 2024 Matthias Nadig

import torch

from ._base import Module


class Add(Module):
    def __init__(self):
        super().__init__()

    def _forward(self, *args):
        x = torch.add(*[arg.data() for arg in args])
        return self._as_batch(data=x, annotations=None, limits=None)
